@@ -27,9 +27,6 @@
 
 #include <glib.h>
 #include <glib/gi18n-lib.h>
-#include <glib-object.h>
-#include <dbus/dbus-glib.h>
-#include <dbus/dbus-glib-lowlevel.h>
 #include <linux/input.h>
 #include <linux/rfkill.h>
 
@@ -41,7 +38,56 @@
 #include "urf-config.h"
 #include "urf-consolekit.h"
 
-#include "urf-daemon-glue.h"
+#define URFKILL_DBUS_INTERFACE "org.freedesktop.URfkill"
+#define URFKILL_OBJECT_PATH "/org/freedesktop/URfkill"
+
+static const char introspection_xml[] =
+"<node>"
+"  <interface name='org.freedesktop.URfkill'>"
+"    <method name='Block'>"
+"      <arg type='u' name='type' direction='in'/>"
+"      <arg type='b' name='block' direction='in'/>"
+"      <arg type='b' name='ret' direction='out'/>"
+"    </method>"
+"    <method name='BlockIdx'>"
+"      <arg type='u' name='index' direction='in'/>"
+"      <arg type='b' name='block' direction='in'/>"
+"      <arg type='b' name='ret' direction='out'/>"
+"    </method>"
+"    <method name='EnumerateDevices'>"
+"      <arg type='ao' name='array' direction='out'/>"
+"    </method>"
+"    <method name='IsInhibited'>"
+"      <arg type='b' name='is_inhibited' direction='out'/>"
+"    </method>"
+"    <method name='Inhibit'>"
+"      <arg type='s' name='reason' direction='in'/>"
+"      <arg type='u' name='inhibit_cookie' direction='out'/>"
+"    </method>"
+"    <method name='Uninhibit'>"
+"      <arg type='u' name='inhibit_cookie' direction='in'/>"
+"    </method>"
+"    <signal name='DeviceAdded'>"
+"      <arg type='o' name='device' direction='out'/>"
+"    </signal>"
+"    <signal name='DeviceRemoved'>"
+"      <arg type='o' name='device' direction='out'/>"
+"    </signal>"
+"    <signal name='DeviceChanged'>"
+"      <arg type='o' name='device' direction='out'/>"
+"    </signal>"
+"    <signal name='UrfkeyPressed'>"
+"      <arg type='i' name='keycode' direction='out'/>"
+"    </signal>"
+"    <property name='DaemonVersion' type='s' access='read'/>"
+"    <property name='KeyControl' type='b' access='read'/>"
+"  </interface>"
+"</node>";
+
+static const GDBusErrorEntry urf_daemon_error_entries[] =
+{
+	{URF_DAEMON_ERROR_GENERAL, "org.freedesktop.URfkill.Daemon.Error.General"},
+};
 
 enum
 {
@@ -65,13 +111,14 @@ static guint signals[SIGNAL_LAST] = { 0 };
 struct UrfDaemonPrivate
 {
 	UrfConfig	*config;
-	DBusGConnection	*connection;
 	UrfPolkit	*polkit;
 	UrfArbitrator   *arbitrator;
 	UrfInput	*input;
 	UrfConsolekit	*consolekit;
 	gboolean	 key_control;
 	gboolean	 master_key;
+	GDBusConnection	*connection;
+	GDBusNodeInfo	*introspection_data;
 };
 
 static void urf_daemon_dispose (GObject *object);
@@ -81,36 +128,6 @@ G_DEFINE_TYPE (UrfDaemon, urf_daemon, G_TYPE_OBJECT)
 #define URF_DAEMON_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), \
 				URF_TYPE_DAEMON, UrfDaemonPrivate))
 
-/**
- * urf_daemon_register_rfkill_daemon:
- **/
-static gboolean
-urf_daemon_register_rfkill_daemon (UrfDaemon *daemon)
-{
-	GError *error = NULL;
-	gboolean ret = FALSE;
-	UrfDaemonPrivate *priv = daemon->priv;
-
-	priv->connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
-	if (priv->connection == NULL) {
-		if (error != NULL) {
-			g_critical ("error getting system bus: %s", error->message);
-			g_error_free (error);
-		}
-		goto out;
-	}
-
-	/* register GObject */
-	dbus_g_connection_register_g_object (priv->connection,
-					     "/org/freedesktop/URfkill",
-					     G_OBJECT (daemon));
-
-	/* success */
-	ret = TRUE;
-out:
-	return ret;
-}
-
 /**
  * urf_daemon_input_event_cb:
  **/
@@ -124,6 +141,7 @@ urf_daemon_input_event_cb (UrfInput *input,
 	UrfArbitrator *arbitrator = priv->arbitrator;
 	gint type;
 	gboolean block = FALSE;
+	GError *error = NULL;
 
 	if (urf_consolekit_is_inhibited (priv->consolekit))
 		goto out;
@@ -169,6 +187,288 @@ urf_daemon_input_event_cb (UrfInput *input,
 	urf_arbitrator_set_block (arbitrator, type, block);
 out:
 	g_signal_emit (daemon, signals[SIGNAL_URFKEY_PRESSED], 0, code);
+	g_dbus_connection_emit_signal (priv->connection,
+	                               NULL,
+	                               URFKILL_OBJECT_PATH,
+	                               URFKILL_DBUS_INTERFACE,
+	                               "UrfkeyPressed",
+	                               g_variant_new ("(i)", code),
+	                               &error);
+	if (error) {
+		g_warning ("Failed to emit UrfkeyPressed: %s", error->message);
+		g_error_free (error);
+	}
+}
+
+/**
+ * urf_daemon_block:
+ **/
+gboolean
+urf_daemon_block (UrfDaemon             *daemon,
+		  const guint            type,
+		  const gboolean         block,
+		  GDBusMethodInvocation *invocation)
+{
+	UrfDaemonPrivate *priv = daemon->priv;
+	PolkitSubject *subject = NULL;
+	gboolean ret = FALSE;
+
+	if (!urf_arbitrator_has_devices (priv->arbitrator))
+		goto out;
+
+	subject = urf_polkit_get_subject (priv->polkit, invocation);
+	if (subject == NULL)
+		goto out;
+
+	if (!urf_polkit_check_auth (priv->polkit, subject, "org.freedesktop.urfkill.block", invocation))
+		goto out;
+
+	ret = urf_arbitrator_set_block (priv->arbitrator, type, block);
+
+	g_dbus_method_invocation_return_value (invocation,
+	                                       g_variant_new ("(b)", ret));
+out:
+	if (subject != NULL)
+		g_object_unref (subject);
+
+	return ret;
+}
+
+/**
+ * urf_daemon_block_idx:
+ **/
+gboolean
+urf_daemon_block_idx (UrfDaemon             *daemon,
+		      const guint            index,
+		      const gboolean         block,
+		      GDBusMethodInvocation *invocation)
+{
+	UrfDaemonPrivate *priv = daemon->priv;
+	PolkitSubject *subject = NULL;
+	gboolean ret = FALSE;
+
+	if (!urf_arbitrator_has_devices (priv->arbitrator))
+		goto out;
+
+	subject = urf_polkit_get_subject (priv->polkit, invocation);
+	if (subject == NULL)
+		goto out;
+
+	if (!urf_polkit_check_auth (priv->polkit, subject, "org.freedesktop.urfkill.blockidx", invocation))
+		goto out;
+
+	ret = urf_arbitrator_set_block_idx (priv->arbitrator, index, block);
+
+	g_dbus_method_invocation_return_value (invocation,
+	                                       g_variant_new ("(b)", ret));
+out:
+	if (subject != NULL)
+		g_object_unref (subject);
+
+	return ret;
+}
+
+/**
+ * urf_daemon_enumerate_devices:
+ **/
+gboolean
+urf_daemon_enumerate_devices (UrfDaemon             *daemon,
+			      GDBusMethodInvocation *invocation)
+{
+	UrfDaemonPrivate *priv = daemon->priv;
+	UrfDevice *device;
+	GVariantBuilder *builder;
+	GVariant *tuple = NULL;
+	GVariant *value = NULL;
+	GList *devices = NULL, *item = NULL;
+
+	g_return_val_if_fail (URF_IS_DAEMON (daemon), FALSE);
+
+	devices = urf_arbitrator_get_devices (priv->arbitrator);
+
+	builder = g_variant_builder_new (G_VARIANT_TYPE("ao"));
+
+	for (item = devices; item; item = g_list_next (item)) {
+		device = (UrfDevice *)item->data;
+		value = g_variant_new_object_path (urf_device_get_object_path (device));
+		g_variant_builder_add_value (builder, value);
+	}
+
+	value = g_variant_builder_end (builder);
+	tuple = g_variant_new_tuple (&value, 1);
+	g_dbus_method_invocation_return_value (invocation, tuple);
+	g_variant_builder_unref (builder);
+	return TRUE;
+}
+
+/**
+ * urf_daemon_is_inhibited:
+ **/
+gboolean
+urf_daemon_is_inhibited (UrfDaemon             *daemon,
+			 GDBusMethodInvocation *invocation)
+{
+	UrfDaemonPrivate *priv = daemon->priv;
+	GVariant *value;
+
+	value = g_variant_new ("(b)", urf_consolekit_is_inhibited (priv->consolekit));
+	g_dbus_method_invocation_return_value (invocation, value);
+
+	return TRUE;
+}
+
+/**
+ * urf_daemon_inhibit:
+ **/
+gboolean
+urf_daemon_inhibit (UrfDaemon             *daemon,
+		    const char            *reason,
+		    GDBusMethodInvocation *invocation)
+{
+	UrfDaemonPrivate *priv = daemon->priv;
+	const char *bus_name;
+	guint cookie = 0;
+
+	bus_name = g_dbus_method_invocation_get_sender (invocation);
+	cookie = urf_consolekit_inhibit (priv->consolekit, bus_name, reason);
+	g_dbus_method_invocation_return_value (invocation,
+	                                       g_variant_new ("(u)", cookie));
+
+	return TRUE;
+}
+
+/**
+ * urf_daemon_uninhibit:
+ **/
+void
+urf_daemon_uninhibit (UrfDaemon             *daemon,
+		      const guint            cookie,
+		      GDBusMethodInvocation *invocation)
+{
+	urf_consolekit_uninhibit (daemon->priv->consolekit, cookie);
+}
+
+static void
+handle_method_call_main (UrfDaemon             *daemon,
+                         const gchar           *method_name,
+                         GVariant              *parameters,
+                         GDBusMethodInvocation *invocation)
+{
+	if (g_strcmp0 (method_name, "Block") == 0) {
+		guint type;
+		gboolean block;
+		g_variant_get (parameters, "(ub)", &type, &block);
+		urf_daemon_block (daemon, type, block, invocation);
+		return;
+	} else if (g_strcmp0 (method_name, "BlockIdx") == 0) {
+		guint index;
+		gboolean block;
+		g_variant_get (parameters, "(ub)", &index, &block);
+		urf_daemon_block_idx (daemon, index, block, invocation);
+		return;
+	} else if (g_strcmp0 (method_name, "EnumerateDevices") == 0) {
+		urf_daemon_enumerate_devices (daemon, invocation);
+		return;
+	} else if (g_strcmp0 (method_name, "IsInhibited") == 0) {
+		urf_daemon_is_inhibited (daemon, invocation);
+		return;
+	} else if (g_strcmp0 (method_name, "Inhibit") == 0) {
+		char *reason;
+		g_variant_get (parameters, "(s)", &reason);
+		urf_daemon_inhibit (daemon, reason, invocation);
+		return;
+	} else if (g_strcmp0 (method_name, "Uninhibit") == 0) {
+		guint cookie;
+		g_variant_get (parameters, "(u)", &cookie);
+		urf_daemon_uninhibit (daemon, cookie, invocation);
+		return;
+	}
+
+	g_assert_not_reached ();
+}
+
+static void
+handle_method_call (GDBusConnection       *connection,
+                    const gchar           *sender,
+                    const gchar           *object_path,
+                    const gchar           *interface_name,
+                    const gchar           *method_name,
+                    GVariant              *parameters,
+                    GDBusMethodInvocation *invocation,
+                    gpointer               user_data)
+{
+	UrfDaemon *daemon = URF_DAEMON (user_data);
+
+	if (g_strcmp0 (interface_name, URFKILL_DBUS_INTERFACE) == 0) {
+		handle_method_call_main (daemon,
+		                         method_name,
+		                         parameters,
+		                         invocation);
+	} else {
+		g_warning ("not recognised interface: %s", interface_name);
+	}
+}
+
+static GVariant *
+handle_get_property (GDBusConnection *connection,
+                     const gchar *sender,
+                     const gchar *object_path,
+                     const gchar *interface_name,
+                     const gchar *property_name,
+                     GError **error,
+                     gpointer user_data)
+{
+	UrfDaemon *daemon = URF_DAEMON (user_data);
+	GVariant *retval = NULL;
+
+	if (g_strcmp0 (property_name, "DaemonVersion") == 0)
+		retval = g_variant_new_string (PACKAGE_VERSION);
+	else if (g_strcmp0 (property_name, "KeyControl") == 0)
+		retval = g_variant_new_boolean (daemon->priv->key_control);
+
+	return retval;
+}
+
+static const GDBusInterfaceVTable interface_vtable =
+{
+	handle_method_call,
+	handle_get_property,
+	NULL, /* handle set_property */
+};
+
+/**
+ * urf_daemon_register_rfkill_daemon:
+ **/
+static gboolean
+urf_daemon_register_rfkill_daemon (UrfDaemon *daemon)
+{
+	UrfDaemonPrivate *priv = daemon->priv;
+	GDBusInterfaceInfo **infos;
+	guint reg_id;
+	GError *error = NULL;
+
+	priv->introspection_data = g_dbus_node_info_new_for_xml (introspection_xml, NULL);
+	g_assert (priv->introspection_data != NULL);
+
+	priv->connection = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, &error);
+	if (priv->connection == NULL) {
+		g_error ("error getting system bus: %s", error->message);
+		g_error_free (error);
+		return FALSE;
+	}
+
+	/* register GObject */
+	infos = priv->introspection_data->interfaces;
+	reg_id = g_dbus_connection_register_object (priv->connection,
+		                                    URFKILL_OBJECT_PATH,
+		                                    infos[0],
+		                                    &interface_vtable,
+		                                    daemon,
+		                                    NULL,
+		                                    NULL);
+	g_assert (reg_id > 0);
+
+	return TRUE;
 }
 
 /**
@@ -213,147 +513,6 @@ out:
 	return ret;
 }
 
-/**
- * urf_daemon_block:
- **/
-gboolean
-urf_daemon_block (UrfDaemon             *daemon,
-		  const guint            type,
-		  const gboolean         block,
-		  DBusGMethodInvocation *context)
-{
-	UrfDaemonPrivate *priv = daemon->priv;
-	PolkitSubject *subject = NULL;
-	gboolean ret = FALSE;
-
-	if (!urf_arbitrator_has_devices (priv->arbitrator))
-		goto out;
-
-	subject = urf_polkit_get_subject (priv->polkit, context);
-	if (subject == NULL)
-		goto out;
-
-	if (!urf_polkit_check_auth (priv->polkit, subject, "org.freedesktop.urfkill.block", context))
-		goto out;
-
-	ret = urf_arbitrator_set_block (priv->arbitrator, type, block);
-
-	dbus_g_method_return (context, ret);
-out:
-	if (subject != NULL)
-		g_object_unref (subject);
-
-	return ret;
-}
-
-/**
- * urf_daemon_block_idx:
- **/
-gboolean
-urf_daemon_block_idx (UrfDaemon             *daemon,
-		      const guint            index,
-		      const gboolean         block,
-		      DBusGMethodInvocation *context)
-{
-	UrfDaemonPrivate *priv = daemon->priv;
-	PolkitSubject *subject = NULL;
-	gboolean ret = FALSE;
-
-	if (!urf_arbitrator_has_devices (priv->arbitrator))
-		goto out;
-
-	subject = urf_polkit_get_subject (priv->polkit, context);
-	if (subject == NULL)
-		goto out;
-
-	if (!urf_polkit_check_auth (priv->polkit, subject, "org.freedesktop.urfkill.blockidx", context))
-		goto out;
-
-	ret = urf_arbitrator_set_block_idx (priv->arbitrator, index, block);
-
-	dbus_g_method_return (context, ret);
-out:
-	if (subject != NULL)
-		g_object_unref (subject);
-
-	return ret;
-}
-
-/**
- * urf_daemon_get_all:
- **/
-gboolean
-urf_daemon_enumerate_devices (UrfDaemon             *daemon,
-			      DBusGMethodInvocation *context)
-{
-	UrfDaemonPrivate *priv = daemon->priv;
-	UrfDevice *device;
-	GPtrArray *object_paths;
-	GList *devices = NULL, *item = NULL;
-
-	g_return_val_if_fail (URF_IS_DAEMON (daemon), FALSE);
-
-	devices = urf_arbitrator_get_devices (priv->arbitrator);
-
-	object_paths = g_ptr_array_sized_new (g_list_length(devices));
-	g_ptr_array_set_free_func (object_paths, g_free);
-	for (item = devices; item; item = g_list_next (item)) {
-		device = (UrfDevice *)item->data;
-		g_ptr_array_add (object_paths, g_strdup (urf_device_get_object_path (device)));
-	}
-
-	dbus_g_method_return (context, object_paths);
-
-	g_ptr_array_unref (object_paths);
-
-	return TRUE;
-}
-
-/**
- * urf_daemon_is_inhibited:
- **/
-gboolean
-urf_daemon_is_inhibited (UrfDaemon             *daemon,
-			 DBusGMethodInvocation *context)
-{
-	UrfDaemonPrivate *priv = daemon->priv;
-
-	dbus_g_method_return (context,
-			      urf_consolekit_is_inhibited (priv->consolekit));
-
-	return TRUE;
-}
-
-/**
- * urf_daemon_inhibit:
- **/
-gboolean
-urf_daemon_inhibit (UrfDaemon             *daemon,
-		    const char            *reason,
-		    DBusGMethodInvocation *context)
-{
-	UrfDaemonPrivate *priv = daemon->priv;
-	char *bus_name;
-	guint cookie = 0;
-
-	bus_name = dbus_g_method_get_sender (context);
-	cookie = urf_consolekit_inhibit (priv->consolekit, bus_name, reason);
-	dbus_g_method_return (context, cookie);
-
-	return TRUE;
-}
-
-/**
- * urf_daemon_uninhibit:
- **/
-void
-urf_daemon_uninhibit (UrfDaemon             *daemon,
-		      const guint            cookie,
-		      DBusGMethodInvocation *context)
-{
-	urf_consolekit_uninhibit (daemon->priv->consolekit, cookie);
-}
-
 /**
  * urf_daemon_device_added_cb:
  **/
@@ -362,6 +521,9 @@ urf_daemon_device_added_cb (UrfArbitrator *arbitrator,
 			    const char    *object_path,
 			    UrfDaemon     *daemon)
 {
+	UrfDaemonPrivate *priv = daemon->priv;
+	GError *error = NULL;
+
 	g_return_if_fail (URF_IS_DAEMON (daemon));
 	g_return_if_fail (URF_IS_ARBITRATOR (arbitrator));
 
@@ -370,6 +532,17 @@ urf_daemon_device_added_cb (UrfArbitrator *arbitrator,
 		return;
 	}
 	g_signal_emit (daemon, signals[SIGNAL_DEVICE_ADDED], 0, object_path);
+	g_dbus_connection_emit_signal (priv->connection,
+	                               NULL,
+	                               URFKILL_OBJECT_PATH,
+	                               URFKILL_DBUS_INTERFACE,
+	                               "DeviceAdded",
+	                               g_variant_new ("(o)", object_path),
+	                               &error);
+	if (error) {
+		g_warning ("Failed to emit DeviceAdded: %s", error->message);
+		g_error_free (error);
+	}
 }
 
 /**
@@ -380,6 +553,9 @@ urf_daemon_device_removed_cb (UrfArbitrator *arbitrator,
 			      const char    *object_path,
 			      UrfDaemon     *daemon)
 {
+	UrfDaemonPrivate *priv = daemon->priv;
+	GError *error = NULL;
+
 	g_return_if_fail (URF_IS_DAEMON (daemon));
 	g_return_if_fail (URF_IS_ARBITRATOR (arbitrator));
 	if (object_path == NULL) {
@@ -387,6 +563,17 @@ urf_daemon_device_removed_cb (UrfArbitrator *arbitrator,
 		return;
 	}
 	g_signal_emit (daemon, signals[SIGNAL_DEVICE_REMOVED], 0, object_path);
+	g_dbus_connection_emit_signal (priv->connection,
+	                               NULL,
+	                               URFKILL_OBJECT_PATH,
+	                               URFKILL_DBUS_INTERFACE,
+	                               "DeviceRemoved",
+	                               g_variant_new ("(o)", object_path),
+	                               &error);
+	if (error) {
+		g_warning ("Failed to emit DeviceRemoved: %s", error->message);
+		g_error_free (error);
+	}
 }
 
 /**
@@ -397,6 +584,9 @@ urf_daemon_device_changed_cb (UrfArbitrator *arbitrator,
 			      const char    *object_path,
 			      UrfDaemon     *daemon)
 {
+	UrfDaemonPrivate *priv = daemon->priv;
+	GError *error = NULL;
+
 	g_return_if_fail (URF_IS_DAEMON (daemon));
 	g_return_if_fail (URF_IS_ARBITRATOR (arbitrator));
 	if (object_path == NULL) {
@@ -404,6 +594,17 @@ urf_daemon_device_changed_cb (UrfArbitrator *arbitrator,
 		return;
 	}
 	g_signal_emit (daemon, signals[SIGNAL_DEVICE_CHANGED], 0, object_path);
+	g_dbus_connection_emit_signal (priv->connection,
+	                               NULL,
+	                               URFKILL_OBJECT_PATH,
+	                               URFKILL_DBUS_INTERFACE,
+	                               "DeviceChanged",
+	                               g_variant_new ("(o)", object_path),
+	                               &error);
+	if (error) {
+		g_warning ("Failed to emit DeviceChanged: %s", error->message);
+		g_error_free (error);
+	}
 }
 
 /**
@@ -460,30 +661,13 @@ urf_daemon_get_property (GObject    *object,
 GQuark
 urf_daemon_error_quark (void)
 {
-	static GQuark ret = 0;
-	if (ret == 0)
-		ret = g_quark_from_static_string ("urf_daemon_error");
-	return ret;
-}
-
-#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }
-/**
- * urf_daemon_error_get_type:
- **/
-GType
-urf_daemon_error_get_type (void)
-{
-	static GType etype = 0;
-
-	if (etype == 0) {
-		static const GEnumValue values[] = {
-			ENUM_ENTRY (URF_DAEMON_ERROR_GENERAL, "GeneralError"),
-			{ 0, 0, 0 }
-		};
-		g_assert (URF_DAEMON_NUM_ERRORS == G_N_ELEMENTS (values) - 1);
-		etype = g_enum_register_static ("UrfDaemonError", values);
-	}
-	return etype;
+	static volatile gsize quark_volatile = 0;
+	g_dbus_error_register_error_domain ("urf_daemon_error",
+	                                     &quark_volatile,
+	                                     urf_daemon_error_entries,
+	                                     G_N_ELEMENTS (urf_daemon_error_entries));
+	G_STATIC_ASSERT (G_N_ELEMENTS (urf_daemon_error_entries) - 1 == URF_DAEMON_ERROR_GENERAL);
+	return (GQuark)quark_volatile;
 }
 
 /**
@@ -545,9 +729,6 @@ urf_daemon_class_init (UrfDaemonClass *klass)
 							       "Whether the key control is enabled",
 							       FALSE,
 							       G_PARAM_READABLE));
-
-	dbus_g_object_type_install_info (URF_TYPE_DAEMON, &dbus_glib_urf_daemon_object_info);
-	dbus_g_error_domain_register (URF_DAEMON_ERROR, NULL, URF_DAEMON_TYPE_ERROR);
 }
 
 /**
@@ -565,7 +746,7 @@ urf_daemon_dispose (GObject *object)
 	}
 
 	if (priv->connection) {
-		dbus_g_connection_unref (priv->connection);
+		g_object_unref (priv->connection);
 		priv->connection = NULL;
 	}
 
